  ' 2048 Game
  ' Version 1.0.0 for CMMD by William M Leue 22-Feb-2048
  ' original game created in 2014 by Gabrielle Cirulli.
  ' adaptation for PicoMite by Volhout 2024

  Option default integer
  Option base 1
  MODE 2
FRAMEBUFFER LAYER 0
FRAMEBUFFER WRITE L


  ' Constants
  Const SIDE   = 4
  Const CSIZE  = 48 '88
  Const BWIDTH = 4  '10
  Const BW     = SIDE*CSIZE
  Const BH     = SIDE*CSIZE
  Const CX     = MM.HRes\2 - BW\2 - (BWIDTH*2)
  Const CY     = MM.VRes\2 - BH\2 - (BWIDTH*2)

  ' Color Indices in Palette
  Dim   map(2)
  Const BCX    = 1
  Const CCX    = 2

  ' Commands
  Const UP    = 128
  Const DOWN  = 129
  Const LEFT  = 130
  Const RIGHT = 131
  Const ESC   = 27
  Const HOME  = 134

  ' Globals
  Dim board(SIDE, SIDE)
  Dim patterns(SIDE*SIDE, SIDE)
  Dim highest_power = 0
  Dim running = 0
  'Matrix
  Font 8
  fw=MM.Info(FONTWIDTH):fh=MM.Info(FONTHEIGHT)
  hh=MM.HRes\fw:vert=MM.VRes\fh
  Dim arr(hh),fade(hh)
  For i=1 To hh: arr(i)=vert*Rnd:fade(i)=&hf*Rnd:Next

  ' Main Prgram
  Open "debug.txt" For output As #1
  MakePallette
  ReadPatterns
  ShowHelp
  InitGame
  HandleEvents
End

  ' Make Color Pallette
Sub MakePallette
  loadcolor
  map(BCX) = colr%(15)
  map(CCX) = colr%(14)
  'map set
End Sub

  ' Read Patterns for Movement
Sub ReadPatterns
  Local i, j
  For i = 1 To SIDE*SIDE
    For j = 1 To SIDE
      Read patterns(i, j)
    Next j
  Next i
End Sub

  ' Initialize a Game
Sub InitGame
  Local row, col
  For row = 1 To SIDE
    For col = 1 To SIDE
      board(row, col) = 0
    Next col
  Next row
  highest_power = 1
  CreateRandomValue
  CreateRandomValue
  running = 1
  CLS : Timer =0
  DrawBoard
End Sub

  ' Create a random 1 or 2 power in a random vacant cell, if one exists.
Sub CreateRandomValue
  If BoardFull() Then Exit Sub
  Do
    row = RandInt(1, SIDE)
    col = RandInt(1, SIDE)
  Loop Until board(row, col) = 0
  board(row, col) = RandInt(1, 2)
  If board(row, col) > highest_power Then highest_power = board(row, col)
End Sub

  ' Draw the Board
Sub DrawBoard
  Local row, col, x, y, s, c, v$
  s = CSIZE + 2*BWIDTH
  s2 = CSIZE + BWIDTH
  For row = 1 To SIDE
    y = CY + (row-1)*s2
    For col = 1 To SIDE
      x = CX + (col-1)*s2
      c = GetCellColor(row, col)
      v$ = GetValue$(row, col)
      DrawCell x, y, s, v$, c
    Next col
  Next row
End Sub

  ' Draw a Cell of the Board
Sub DrawCell x, y, s, v$, c
  Local nx, ny
  'Box x, y, s, s,, map(BCX), map(BCX)
  RBox x+BWIDTH, y+BWIDTH, CSIZE, CSIZE,8, map(BCX), c
  nx = x + BWIDTH+CSIZE\2
  ny = y + BWIDTH+CSIZE\2
  If Len(v$) > 0 Then
    Text nx, ny, v$, "CM", 7,, map(BCX), -1
    'text nx+1, ny, v$, "CM", 2,, rgb(black), -1
  EndIf
End Sub

  ' Assign a color to a board cell depending on its current value
Function GetCellColor(row, col)
  GetCellColor = map(CCX)
  p = board(row, col)
  If p > 0 Then
    GetCellColor = colr%(p)
  EndIf
End Function

  ' return the string numeric value of a board cell
  ' if that cell's power >= 1, else return the empty string.
Function GetValue$(row, col)
  Local nx, ny, p, v
  p = board(row, col)
  If p > 0 Then
    v = 2^p
    GetValue$ = Str$(v)
  Else
    GetValue$ = ""
  EndIf
End Function

  ' Get User Keyboard Inputs
Sub HandleEvents
  Local z$, cmd
  z$ = Inkey$
  Do
    Do
      Matrix
      z$ = Inkey$
      If Timer > 20000 Then ScreenSaver:Timer =0:DrawBoard
    Loop Until z$ <> ""
    Timer =0
    cmd = Asc(UCase$(z$))
    If cmd = ESC Then
      CLS
      End
    EndIf
    If cmd = HOME Then
      InitGame
    EndIf
    If running Then
      Select Case cmd
        Case UP
          Merge(UP)
        Case DOWN
          Merge(DOWN)
        Case LEFT
          Merge(LEFT)
        Case RIGHT
          Merge(RIGHT)
      End Select
      DrawBoard
      Play Tone 100,100,50
    EndIf
    If GameOver() Then
      running = 0
      ShowGameOver
    EndIf
  Loop
End Sub

  ' Shift Values and Merge in the specified direction
Sub Merge cmd
  Local row, col
  Select Case cmd
    Case UP
      PushCellsUp
    Case DOWN
      PushCellsDown
    Case LEFT
      PushCellsLeft
    Case RIGHT
      PushCellsRight
  End Select
  CreateRandomValue
End Sub

  ' Push cells up and merge
Sub PushCellsUp
  Local row, col, p
  Local pat(SIDE)
  For col = 1 To SIDE
    row = 1
    MatchPattern row, col, 1, 0, pat()
    For row = 1 To SIDE
      p = pat(row)
      If p > 0 Then
        board(row-p, col) = board(row, col)
        board(row, col) = 0
      EndIf
    Next row
    MergeUp col
  Next col
End Sub

  ' Merge matching cells upwards
Sub MergeUp col
  Local row, nrow
  For row = 1 To SIDE-1
    If board(row, col) > 0 Then
      If board(row+1, col) = board(row, col) Then
        Inc board(row, col)
        If board(row, col) > highest_power Then highest_power = board(row, col)
        For nrow = row+2 To SIDE
          board(nrow-1, col) = board(nrow, col)
        Next nrow
        board(SIDE, col) = 0
      EndIf
    EndIf
  Next row
End Sub

  ' Push cells down and merge
Sub PushCellsDown
  Local row, col, p
  Local pat(SIDE)
  For col = 1 To SIDE
    row = SIDE
    MatchPattern row, col, -1, 0, pat()
    For row = SIDE To 1 Step -1
      p = pat(row)
      If p > 0 Then
        board(row+p, col) = board(row, col)
        board(row, col) = 0
      EndIf
    Next row
    MergeDown col
  Next col
End Sub

  ' Merge matching cells downwards
Sub MergeDown col
  Local row, nrow
  For row = SIDE To 2 Step -1
    If board(row, col) > 0 Then
      If board(row-1, col) = board(row, col) Then
        Inc board(row, col)
        If board(row, col) > highest_power Then highest_power = board(row, col)
        For nrow = row-2 To 1 Step -1
          board(nrow+1, col) = board(nrow, col)
        Next nrow
        board(1, col) = 0
      EndIf
    EndIf
  Next row
End Sub

  ' Push cells Left and merge
Sub PushCellsLeft
  Local row, col, p
  Local pat(SIDE)
  For row = 1 To SIDE
    col = 1
    MatchPattern row, col, 0, 1, pat()
    For col = 1 To SIDE
      p = pat(col)
      If p > 0 Then
        board(row, col-p) = board(row, col)
        board(row, col) = 0
      EndIf
    Next col
    MergeLeft row
  Next row
End Sub

  ' Merge matching cells leftwards
Sub MergeLeft row
  Local col, ncol
  For col = 1 To SIDE-1
    If board(row, col) > 0 Then
      If board(row, col+1) = board(row, col) Then
        Inc board(row, col)
        If board(row, col) > highest_power Then highest_power = board(row, col)
        For ncol = col+2 To SIDE
          board(row, ncol-1) = board(row, ncol)
        Next ncol
        board(row, SIDE) = 0
      EndIf
    EndIf
  Next col
End Sub

  ' Push cells right and merge
Sub PushCellsRight
  Local row, col, p
  Local pat(SIDE)
  For row = 1 To SIDE
    col = SIDE
    MatchPattern row, col, 0, -1, pat()
    For col = SIDE To 1 Step -1
      p = pat(col)
      If p > 0 Then
        board(row, col+p) = board(row, col)
        board(row, col) = 0
      EndIf
    Next col
    MergeRight row
  Next row
End Sub

  ' Merge matching cells rightwards
Sub MergeRight row
  Local col, ncol
  For col = SIDE To 2 Step -1
    If board(row, col) > 0 Then
      If board(row, col-1) = board(row, col) Then
        Inc board(row, col)
        If board(row, col) > highest_power Then highest_power = board(row, col)
        For ncol = col-2 To 1 Step -1
          board(row, ncol+1) = board(row, ncol)
        Next ncol
        board(row, 1) = 0
      EndIf
    EndIf
  Next col
End Sub

  ' Find the pattern of values that matches a template.
  ' We add up the non-zero cells not by their values, but
  ' by the powers of 2 that match their location. This yields
  ' an index into the template array which gives the spaces
  ' to be shifted during a move in any direction.
Sub MatchPattern row, col, dr, dc, pat()
  Local i, index, p = 3, crow, ccol, px
  index = 0 : crow = row : ccol = col
  For i = 1 To SIDE
    If board(crow, ccol) > 0 Then
      Inc index, 2^p
    EndIf
    Inc crow, dr : Inc ccol, dc : Inc p, -1
  Next i
  Inc index, 1
  If dr <> 0 Then
    px = row
  Else
    px = col
  EndIf
  For i= 1 To 4
    pat(px) = patterns(index, i)
    If dr <> 0 Then
      Inc px, dr
    Else
      Inc px, dc
    EndIf
  Next i
End Sub

  ' Returns 1 if board is full or zero if not
Function BoardFull()
  Local row, col
  For row = 1 To SIDE
    For col = 1 To SIDE
      If board(row, col) = 0 Then
        BoardFull = 0
        Exit Function
      EndIf
    Next col
  Next row
  BoardFull = 1
End Function

  ' The game is over when all cells have non-zero values and
  ' no cells can be merged
Function GameOver()
  Local row, col, p1, p2, f
  GameOver = 0
  f = BoardFull()
  If f Then
    For row = 1 To SIDE
      For col = 1 To SIDE-1
        p1 = board(row, col)
        p2 = board(row, col+1)
        If p1 = p2 Then Exit Function
      Next col
    Next row
    For col = 1 To SIDE
      For row = 1 To SIDE-1
        p1 = board(row, col)
        p2 = board(row+1, col)
        If p1 = p2 Then Exit Function
      Next row
    Next col
    GameOver = 1
  EndIf
End Function

  ' Show the Game Over
Sub ShowGameOver
  Local m$
  Text MM.HRes\2, CSIZE, "GAME OVER", "CT", 4,3, RGB(cyan), -1
  m$ = "Highest Power of 2 Achieved: " + Str$(2^highest_power)
  Text MM.HRes\2, 2, m$, "CT", 4,, RGB(red), -1
  If highest_power >= 11 Then
    Text MM.HRes\2, 60, "*** WINNER ***", "CT", 4,, RGB(red), -1
  EndIf
  Text MM.HRes\2, MM.VRes-2, "Press Home Key for a New Game", "CB", 4,, RGB(yellow), -1
End Sub

  ' Offer Help
Sub ShowHelp
  Text MM.HRes\2, 0, "Rules for the 2048 Game", "CT", 7,2, RGB(cyan)
  Colour RGB(green)
  Font 8
  Print @(0, 20)
  Print " The goal of the 2048 game is to merge cells on a 4x4 board until you get to"
  Print "the value of 2048 (or up to 131,072, which is the highest possible value)."
  Print "The board starts out with all cells empty except for 2 random cells, which"
  Print "have the values of 2 or 4, randomly chosen. Use the arrow keys to slide cells"
  Print "up,down,left and right. All the cells on the board will be slid as far as"
  Print "possible in the chosen direction."
  Print ""
  Print " Each time you slide cells in a direction, any two cells which are adjacent in"
  Print "that direction and which have the same value will be merged into a single cell"
  Print "that has the sum of the values of the two cells. So for instance, if you press"
  Print "the LEFT arrow and there are two adjacent cells in a row, each having a 2 in"
  Print "them, they will be merged into a single cell having a 4 in it."
  Print ""
  Print " If there are 3 adjacent cells having the same value, only the two cells"
  Print "closest to the edge that the cells are sliding towards will get merged. But"
  Print "if there are 4 adjacent cells having the same value, then each pair of cells"
  Print "will be merged. However, theresulting pair of identical cells of the next"
  Print "higher value will require another key press to merge."
  Print ""
  Print " After each move, an empty cell will be randomly filled with a 2 or a 4."
  Print ""
  Print " The game is over when all the cells are filled or there are no more adjacent"
  Print "cells that can be merged with a move in any direction."
  Print ""
  Print " You can press the HOME key at any time to start a new game. To quit the"
  Print "program press the ESCAPE key."
  Print ""
  Print " The 2048 game was invented by Gabriele Cirulli in 2014. This version for the"
  Print "CMM2 was reverse-engineered by William M. Leue using the Wikipedia"
  Print "description of the game."
  Text MM.HRes\2, MM.VRes-20, "Press Any Key to Continue","CT",7,,RGB(yellow)
  Do
  Loop Until Inkey$ <> ""
'  CLS
End Sub

  ' return a uniformly distributed random integer in the specified closed range
Function RandInt(a As integer, b As integer)
  Local integer v, c
  c = b-a+1
  Do
    v = a + (b-a+2)*Rnd()
    If v >= a And v <= b Then Exit
  Loop
  RandInt = v
End Function

  ' Cell Population templates for shifting.
  Data 0, 0, 0, 0
  Data 0, 0, 0, 3
  Data 0, 0, 2, 0
  Data 0, 0, 2, 2
  Data 0, 1, 0, 0
  Data 0, 1, 0, 2
  Data 0, 1, 1, 0
  Data 0, 1, 1, 1
  Data 0, 0, 0, 0
  Data 0, 0, 0, 2
  Data 0, 0, 1, 0
  Data 0, 0, 1, 1
  Data 0, 0, 0, 0
  Data 0, 0, 0, 1
  Data 0, 0, 0, 0
  Data 0, 0, 0, 0

Sub loadcolor
  Dim colr%(16)
  colr%(1)=RGB(white)
  colr%(2)=RGB(yellow)
  colr%(3)=RGB(lilac)
  colr%(4)=RGB(brown)
  colr%(5)=RGB(fuchsia)
  colr%(6)=RGB(rust)
  colr%(7)=RGB(magenta)
  colr%(8)=RGB(red)
  colr%(9)=RGB(cyan)
  colr%(10)=RGB(green)
  colr%(11)=RGB(cerulean)
  colr%(12)=RGB(midgreen)
  colr%(13)=RGB(cobalt)
  colr%(14)=RGB(myrtle)
  colr%(15)=RGB(blue)
  colr%(16)=RGB(black)
End Sub

Sub ScreenSaver
Font 8
'Local fw=MM.Info(FONTWIDTH):fh=MM.Info(FONTHEIGHT)
'Local hh=MM.HRes\fw:vert=MM.VRes\fh
'Local arr(hh),fade(hh)
'Local i
'For i=1 To hh: arr(i)=vert*Rnd:fade(i)=&hf*Rnd:Next
Local c,x
Do
For x = 1 To hh
  c=&h1000*(fade(x)-&hf) And &hff00
  Text x*fw-fw,arr(x)*fh,Chr$(Rnd*75+48),,,,c
  If arr(x)>vert Then arr(x)=0 Else Inc arr(x)
  If fade(x) = &hf0 Then fade(x)=0 Else Inc fade(x)
Next x
Pause 100
Loop Until Inkey$ <> ""
CLS
End Sub

Sub Matrix
Font 8
FRAMEBUFFER WRITE N
Local c,x
For x = 1 To hh
  c=&h1000*(fade(x)-&hf) And &hff00
  Text x*fw-fw,arr(x)*fh,Chr$(Rnd*75+48),,,,c
  If arr(x)>vert Then arr(x)=0 Else Inc arr(x)
  If fade(x) = &hf0 Then fade(x)=0 Else Inc fade(x)
Next x
FRAMEBUFFER WRITE L
End Sub
